#include <stdio.h>
#include <string.h>
#include <curses.h>
#include <math.h>
#include "main.h"       // our own stuff


  //
  //             R-20 SIMULATOR
  //
  //      jul/aug 2012, insektoiden.de


  // "curses" stuff
  WINDOW *wnd;
  int nrows;
  int ncols;


  // data representation

  // How to represent the 4 registers?
  // Originally, there were 12 nibbles for BCD digits, 2 nibbles for
  // BCD coded exponent and 1 nibble each for the signs.
  //
  //   +-   +-    1 ,  2    3    4    5    6    7    8    9    0    1    2
  // |....|....|....|....|....|....|....|....|....|....|....|....|....|....|
  //
  //  sign sign  1G  100M  10M   1M 100k  10k  1k   100  10    1   10    1
  //  man  exp  \_________________ man _________________________/ \__exp__/
  //
  // |    0    |    1    |    2    |    3    |    4    |    5    |    6    |
  //
  //
  // sign     [0001] or [1111],  two nibbles       Byte   0
  // mantissa [0..9] ... [0..9], ten bcd digits    Bytes 1-5
  // exponent [0..9] + [0..9],   two bcd digits    Byte   6


#define SIGNS 0
#define MAN   1
#define EXP   6

#define X 0
#define Y 1
#define Z 2
#define T 3
#define A 4
#define B 5

  __uint8_t reg[6][7];




/**
 * Convert register to string.
 */
 // "--" "1.234567890" "12" --> "-1.234567890e-12"
 //
void rtostr(char str[], __uint8_t* r) {

  // check for man sign coded in Byte 0
  if ((r[SIGNS] & 0xf0) == 0x10) str[0] = ' ';
  else if ((r[SIGNS] & 0xf0) == 0xf0) str[0] = '-';
  else str[0] = '*';

  // access first digit of mantissa
  str[1] = ((r[MAN] & 0xf0)>>4) + '0';
  str[2] = '.';

  // access second digit of mantissa
  str[3] = (r[MAN] & 0x0f) + '0';

  // access digit 3-10 of mantissa
  int i;
  for (i=0; i<=3; i++) {
    str[4+2*i] = ((r[MAN+1+i] & 0xf0)>>4) + '0';
    str[5+2*i] =  (r[MAN+1+i] & 0x0f) + '0';
  }
  str[12] = 'e';

  // check for exp sign coded in Byte 0
  if ((r[SIGNS] & 0x0f) == 0x01) str[13] = ' ';
  else if ((r[SIGNS] & 0x0f) == 0x0f) str[13] = '-';
  else str[13] = '*';

  // access digits of exponent
  str[14] = ((r[EXP] & 0xf0)>>4) + '0';
  str[15] =  (r[EXP] & 0x0f) + '0';
  str[16] = '\0';
}

/**
 * Clear screen and print values of t,z,y,x registers.
 */
void screen_refresh() {
  char str[17];   // will hold number, e.g. "-1.234567890e-12\0"

//  clear();
  move(0,0);
  printw("+------------------+\n");
  printw("| ");
  rtostr(str, reg[T]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[Z]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[Y]);
  printw(str);
  printw(" |\n");
  printw("| ");
  rtostr(str, reg[X]);
  printw(str);
  printw(" |\n");
  printw("+------------------+");

  move(nrows-1, ncols-1);       // "hide" cursor
} // screen_refresh


void err_error() {
  move(4, 0);
  printw("|  invalid input     |\n");
  move(nrows-1, ncols-1);       // "hide" cursor
  getch();
}

/**
 * Copy contents of src to dst, whose content will be lost.
 */
void copy_reg(__int8_t* dst, __int8_t* src) {
  memcpy(dst, src, 7);
}

/**
 * Return exponent of register.
 */
int get_exp(__int8_t* r) {
  int e;

  e = (r[EXP]>>4)*10 + (r[EXP] & 0x0f);
  if ((r[SIGNS] & 0x0f) == 0x0f) e = -e;
  return e;
}

/**
 * Shift mantissa 1 digit to the left.
 */
void shiftl1(__int8_t* r) {
//  exp_inc(r);
  r[MAN+4] = (r[MAN+4] >> 4) + ((r[MAN+3]&0x0f)<<4);
  r[MAN+3] = (r[MAN+3] >> 4) + ((r[MAN+2]&0x0f)<<4);
  r[MAN+2] = (r[MAN+2] >> 4) + ((r[MAN+1]&0x0f)<<4);
  r[MAN+1] = (r[MAN+1] >> 4) + ((r[MAN]  &0x0f)<<4);
  r[MAN]   = (r[MAN]&0xf0) >> 4;
}

/**
 * Shift mantissa n digits to the left.
 */
void shiftl(__int8_t* r, int n) {
 // for (; n > 0; n--)
 shiftl1(r);

}

/*
fn_sin() {
  x = sin(x);
}

fn_asin() {
  if (x >= -1 && x <= 1) x = asin(x);
  else err_error();
}

fn_digit(int digit) {
  switch(state) {
    case INT_PART: {
      if (nr_digits < 10) {
        x*=10;
        x+=digit;
        nr_digits++;
      }
      break;
    }
    case DEC_PART: {
      if (nr_digits < 10) display_dec++;
    }
    case EXP_PART {
      if (nr_exp_digits < 3) {
        exp_x*=10;
        nr_exp_digits++;
      }
    }
  } // switch
}
//*/

fn_pi() {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);

  reg[X][SIGNS] = 0x11;
  reg[X][MAN]   = 0x31;
  reg[X][MAN+1] = 0x41;
  reg[X][MAN+2] = 0x59;
  reg[X][MAN+3] = 0x26;
  reg[X][MAN+4] = 0x54;
  reg[X][EXP]   = 0x00;
}

fn_plus() {
  int ex;
  int ey;

  ex = get_exp(reg[X]);
  ey = get_exp(reg[Y]);

  if (ex > ey) shiftl(reg[Y], ex-ey);
  else if (ey > ex) shiftl(reg[X], ey-ex);

}

fn_enter() {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);
}

void call(int key) {
  switch(key) {
//    case 's': { fn_sin(); break; }
//    case 'a': { fn_asin(); break; }
    case 'p': { fn_pi(); break; }
//    case '0': { fn_digit(0); break; }
//    case '1': { fn_digit(1); break; }
//    case '2': { fn_digit(2); break; }
//    case '3': { fn_digit(3); break; }
//    case '4': { fn_digit(4); break; }
//    case '5': { fn_digit(5); break; }
//    case '6': { fn_digit(6); break; }
//    case '7': { fn_digit(7); break; }
//    case '8': { fn_digit(8); break; }
//    case '9': { fn_digit(9); break; }
    case 10: { fn_enter(); break; }
    case '+': { fn_plus(); break; }
  }
}


void init_regs() {
  //example
  reg[T][SIGNS] = 0x11;
  reg[T][MAN]   = 0x31;
  reg[T][MAN+1] = 0x41;
  reg[T][MAN+2] = 0x59;
  reg[T][MAN+3] = 0x26;
  reg[T][MAN+4] = 0x54;
  reg[T][EXP]   = 0x00;

  reg[Z][SIGNS] = 0xff;
  reg[Z][MAN]   = 0x27;
  reg[Z][MAN+1] = 0x18;
  reg[Z][MAN+2] = 0x48;
  reg[Z][MAN+3] = 0x18;
  reg[Z][MAN+4] = 0x48;
  reg[Z][EXP]   = 0x23;

  reg[Y][SIGNS] = 0x1f;
  reg[Y][MAN]   = 0x99;
  reg[Y][MAN+1] = 0x37;
  reg[Y][MAN+2] = 0x25;
  reg[Y][MAN+3] = 0x28;
  reg[Y][MAN+4] = 0x31;
  reg[Y][EXP]   = 0x05;

  reg[X][SIGNS] = 0xf1;
  reg[X][MAN]   = 0x12;
  reg[X][MAN+1] = 0x34;
  reg[X][MAN+2] = 0x56;
  reg[X][MAN+3] = 0x78;
  reg[X][MAN+4] = 0x90;
  reg[X][EXP]   = 0x12;
} // init_regs


int main(void) {

/*printf("(0x99 & 0xf0) /16 = 0x%x \n", (0x99 & 0xf0) /16);*/
/*exit(-1);*/

  char stri[32];

  wnd = initscr();                // initialize curses
  cbreak();                       // send keys directly to program
  noecho();                       // don't write my typing to screen
  getmaxyx(wnd, nrows, ncols);    // check terminal dimensions

  init_regs();

  int key = 0;
  do {
    screen_refresh();
    key = getch();
move(12,0);
printw("              ");
move(12,0);
printw("Taste: %i", key);
    call(key);
  } while (key != 27);            // 27 = escape (ASCII)

  echo();
  endwin();
  return 0;

} // main
